From 1b06231da8cd213e30c4fbe5115bfc87c8873754 Mon Sep 17 00:00:00 2001
From: Jimmy Yih <jyih@vmware.com>
Date: Wed, 9 Aug 2023 16:50:04 -0700
Subject: [PATCH] Archive current timeline history file after recovery finishes
 if needed

Previously, the only time the timeline history file would be archived
was when a standby already configured for WAL archiving was
promoted. If WAL archiving was set up and started after the standby
was promoted, its current timeline history file would not be found in
the archive. This could cause issues when restoring backups that did
not include the timeline history files (e.g. pg_basebackup
-Xnone). Some example failures include failing to restore with
recovery_target_timeline explicitly set to the control file's timeline
id and failing to create cascade standbys after recovering with
recovery_target_timeline set to 'current' or 'latest'. To prevent
these restore issues, we now ensure that the current timeline history
file has been archived by marking it as ready for archiving after
recovery finishes if it was found that it has not already been
archived or marked ready.
---
 src/backend/access/transam/xlog.c             |  31 ++++
 src/backend/access/transam/xlogarchive.c      |   6 +-
 .../t/038_archive_current_timeline_history.pl | 134 ++++++++++++++++++
 3 files changed, 168 insertions(+), 3 deletions(-)
 create mode 100644 src/test/recovery/t/038_archive_current_timeline_history.pl

diff --git a/src/backend/access/transam/xlog.c b/src/backend/access/transam/xlog.c
index f6f8adc72a..99b53b9a7e 100644
--- a/src/backend/access/transam/xlog.c
+++ b/src/backend/access/transam/xlog.c
@@ -5800,6 +5800,37 @@ StartupXLOG(void)
 	if (standbyState != STANDBY_DISABLED)
 		ShutdownRecoveryTransactionEnvironment();
 
+	/*
+	 * It's possible that this could be the first time WAL archiving has been
+	 * enabled.  If the timeline is greater than 1, we need to check if the
+	 * current timeline history file needs to be archived.  This will prevent
+	 * any PITR-related issues later on where a timeline history file is
+	 * required.
+	 */
+	if (XLogArchivingActive())
+	{
+		TimeLineID	currentTimeLineID;
+
+		currentTimeLineID = GetWALInsertionTimeLine();
+		if (currentTimeLineID > 1) {
+			char	histfname[MAXFNAMELEN];
+
+			/*
+			 * Timeline history .done files do not get removed automatically
+			 * so this check should be valid to make sure we don't archive the
+			 * timeline history file again on restart.  However, if the
+			 * timeline history .done file was manually removed for some
+			 * reason, then we make the assumption that the archive_command is
+			 * set up properly to gracefully handle the re-archiving attempt.
+			 * If there's already a .ready or .done file, then there's nothing
+			 * to do.
+			 */
+			TLHistoryFileName(histfname, currentTimeLineID);
+			if (!XLogArchiveIsReadyOrDone(histfname))
+				XLogArchiveNotify(histfname);
+		}
+	}
+
 	/*
 	 * If there were cascading standby servers connected to us, nudge any wal
 	 * sender processes to notice that we've been promoted.
diff --git a/src/backend/access/transam/xlogarchive.c b/src/backend/access/transam/xlogarchive.c
index f3fb92c8f9..dee679bd18 100644
--- a/src/backend/access/transam/xlogarchive.c
+++ b/src/backend/access/transam/xlogarchive.c
@@ -650,9 +650,9 @@ XLogArchiveIsBusy(const char *xlog)
  * This is similar to XLogArchiveIsBusy(), but returns true if the file
  * is already archived or is about to be archived.
  *
- * This is currently only used at recovery.  During normal operation this
- * would be racy: the file might get removed or marked with .ready as we're
- * checking it, or immediately after we return.
+ * This is primarily used at recovery.  During normal operation this would be
+ * racy: the file might get removed or marked with .ready as we're checking
+ * it, or immediately after we return.
  */
 bool
 XLogArchiveIsReadyOrDone(const char *xlog)
diff --git a/src/test/recovery/t/038_archive_current_timeline_history.pl b/src/test/recovery/t/038_archive_current_timeline_history.pl
new file mode 100644
index 0000000000..41e2b5b2d4
--- /dev/null
+++ b/src/test/recovery/t/038_archive_current_timeline_history.pl
@@ -0,0 +1,134 @@
+# Test that setting up and starting WAL archiving on an
+# already-promoted node will result in the archival of its current
+# timeline history file.
+
+use strict;
+use warnings;
+use PostgreSQL::Test::Cluster;
+use PostgreSQL::Test::Utils;
+use Test::More;
+
+$ENV{PGDATABASE} = 'postgres';
+
+# Initialize primary node
+my $node_primary = PostgreSQL::Test::Cluster->new('primary');
+$node_primary->init(allows_streaming => 1);
+$node_primary->start;
+
+# Take a backup
+my $backup_name = 'my_backup_1';
+$node_primary->backup($backup_name);
+
+# Create a standby that will be promoted onto timeline 2
+my $node_primary_tli2 = PostgreSQL::Test::Cluster->new('primary_tli2');
+$node_primary_tli2->init_from_backup($node_primary, $backup_name,
+	has_streaming => 1);
+$node_primary_tli2->start;
+
+# Stop and remove the primary; it's not needed anymore
+$node_primary->teardown_node;
+
+# Promote the standby using "pg_promote", switching it to timeline 2
+my $psql_out = '';
+$node_primary_tli2->psql(
+	'postgres',
+	"SELECT pg_promote(wait_seconds => 300);",
+	stdout => \$psql_out);
+is($psql_out, 't', "promotion of standby with pg_promote");
+
+# Enable archiving on the promoted node. The timeline 2 history file
+# will be pushed to the archive.
+$node_primary_tli2->enable_archiving;
+$node_primary_tli2->restart;
+
+# Check that the timeline 2 history file has been archived. The file
+# is marked ready and immediately archived when the archiver process
+# starts up but loop to make this check deterministic.
+my $primary_archive = $node_primary_tli2->archive_dir;
+$node_primary_tli2->poll_query_until('postgres',
+	"SELECT size IS NOT NULL FROM pg_stat_file('$primary_archive/00000002.history', true);")
+  or die "Timed out while waiting for 00000002.history to be archived";
+
+# Take backup of node_primary_tli2 and use -Xnone so that pg_wal will
+# be empty and restore will retrieve the necessary WAL and timeline
+# history file(s) from the archive.
+$backup_name = 'my_backup_2';
+$node_primary_tli2->backup($backup_name, backup_options => ['-Xnone']);
+
+# Create simple WAL that will be archived and restored
+$node_primary_tli2->safe_psql('postgres', "CREATE TABLE tab_int AS SELECT 8 AS a;");
+
+# Create a restore point to later use as the recovery_target_name
+my $recovery_name = "my_target";
+$node_primary_tli2->safe_psql('postgres',
+	"SELECT pg_create_restore_point('$recovery_name');");
+
+# Find the next WAL segment to be archived
+my $walfile_to_be_archived = $node_primary_tli2->safe_psql('postgres',
+	"SELECT pg_walfile_name(pg_current_wal_lsn());");
+
+# Make the WAL segment eligible for archival
+$node_primary_tli2->safe_psql('postgres', 'SELECT pg_switch_wal();');
+
+# Wait until the WAL segment has been archived
+my $archive_wait_query =
+  "SELECT '$walfile_to_be_archived' <= last_archived_wal FROM pg_stat_archiver;";
+$node_primary_tli2->poll_query_until('postgres', $archive_wait_query)
+  or die "Timed out while waiting for WAL segment to be archived";
+$node_primary_tli2->teardown_node;
+
+# Initialize a new standby node from the backup. This node will start
+# off on timeline 2 according to the control file and will finish
+# recovery onto the same timeline by explicitly setting
+# recovery_target_timeline to '2'. We explicitly set the timeline id
+# because startup will fail if the timeline history file is not
+# retrievable from the archive but will not fail if we use 'current'
+# or 'latest'.
+my $node_standby = PostgreSQL::Test::Cluster->new('standby');
+$node_standby->init_from_backup($node_primary_tli2, $backup_name,
+	has_restoring => 1, standby => 0);
+$node_standby->append_conf('postgresql.conf', qq{
+recovery_target_timeline = '2'
+recovery_target_action = 'pause'
+recovery_target_name = 'my_target'
+archive_mode = 'off'
+primary_conninfo = ''
+});
+$node_standby->start;
+
+# Sanity check that the timeline history file was retrieved
+my $log_contents = slurp_file($node_standby->logfile);
+ok( $log_contents =~ qr/restored log file "00000002.history" from archive/,
+	"00000002.history retrieved from the archives");
+ok ( -f $node_standby->data_dir . "/pg_wal/00000002.history",
+  "00000002.history exists in the standby's pg_wal directory");
+
+# Sanity check that the node is queryable
+my $result_standby =
+  $node_standby->safe_psql('postgres', "SELECT timeline_id FROM pg_control_checkpoint();");
+is($result_standby, qq(2), 'check that the node is on timeline 2');
+$result_standby =
+  $node_standby->safe_psql('postgres', "SELECT * FROM tab_int;");
+is($result_standby, qq(8), 'check that the node did archive recovery');
+
+# Set up a cascade standby node to validate that there's no issues
+# since the WAL receiver will request all necessary timeline history
+# files from the standby node's WAL sender.
+my $node_cascade = PostgreSQL::Test::Cluster->new('cascade');
+$node_cascade->init_from_backup($node_primary_tli2, $backup_name,
+	standby => 1);
+$node_cascade->enable_streaming($node_standby);
+$node_cascade->start;
+
+# Wait for the replication to catch up
+$node_standby->wait_for_catchup($node_cascade);
+
+# Sanity check that the cascade standby node came up and is queryable
+my $result_cascade =
+  $node_cascade->safe_psql('postgres', "SELECT * FROM tab_int;");
+is($result_cascade, qq(8), 'check that the node received the streamed WAL data');
+
+$node_standby->teardown_node;
+$node_cascade->teardown_node;
+
+done_testing();
-- 
2.24.3 (Apple Git-128)

